#!/usr/bin/perl -w
use Net::Pcap;
use NetPacket::Ethernet qw(:strip);
use NetPacket::IP qw(:strip);
use NetPacket::TCP;
use NetPacket::ICMP;

use IO::Handle;

use Net::DNS;
use Getopt::Long;

$|=1;

use strict;

my %legal = (
5900 => 1,
7557 => 1,
1433 => 1,
20 => 1,
23790 => 1,
34357 => 1,
35999 => 1,
37370 => 1,
16818 => 1,
25 => 1,
37713 => 1,
49110 => 1,
ICMP => 1,
UNKNOWN_PORT => 1,
UNKNOWN_PROTOCOL => 1
);


my $all="";
my $result = GetOptions("all" => \$all);

# directly process file...
my $fn = shift @ARGV; my $err=undef;
my $capt = Net::Pcap::open_offline($fn,\$err);
if (!$capt) { print $err,"\n"; exit; }

open(F,">> LOG.trackPackets");
F->autoflush(1);

# read in all previous packets
my $res = Net::DNS::Resolver->new( nameservers => [qw(213.208.131.102)], tcp_timeout => 10, udp_timeout => 10 );

my @is=();
my %dnscache;
if ($all) {
  Net::Pcap::loop($capt, -1, \&process_pkt, "TCP");
} else {
  Net::Pcap::loop($capt, -1, \&process_pkt_ignore, "TCP");
}

$SIG{"INT"} = sub { Net::Pcap::close($capt); die "Closed connection"; };

# new capture file name
my $id = length($fn)-1;
while (substr($fn,$id,1) =~ /^[0-9]$/) { $id--; }
my $fn2=substr($fn,0,$id+1);
my $k;
if (substr($fn,$id+1) eq "") { $k=1; } else {
  $k = substr($fn,$id+1)+1;
}
$fn2.=$k;

open(G,"< train.arff");
my $header=""; my $inH=1;
while (<G>) {
  if ($inH) {
    $header.=$_;
    if (/^\@data/) { $inH=undef; }
  }
}
close(G);

while (! -e $fn2) {
  %dnscache=();
  Net::Pcap::dispatch($capt, 0, \&process_pkt, "TCP");

	if (keys %dnscache == 0) { sleep(55); next; }

	open(G,"> $$.test.arff");
	print G $header;
	foreach my $k (@is) {
		print G join(",",@{$k->[0]}),"\n";
	}
	close(G);

	system("java -mx2048M weka.filters.unsupervised.attribute.StringToWordVector -C -b -i train.arff -o train-wv.arff -r $$.test.arff -s $$.test-wv.arff -tokenizer \'weka.core.tokenizers.WordTokenizer -delimiters \",\"\'");

	open(G,"java -mx2048M weka.classifiers.functions.SMO -o -p 0 -t train-wv.arff -T $$.test-wv.arff -c 5 |");

	$_=<G>; chomp; $_ eq " inst#     actual  predicted error prediction" or die "Format error, have you changed WEKA version?";
	my $i=0;
	while (<G>) {
  	chomp;
		s/[ \+][ \+]*/ /g;
		s/^[ \t]//g;
		if ($_ eq "") { next; }
		my @x=split(/ /,$_);
		$x[0]-1 == $i or die "Order mismatch";
	
		my ($cl,$tmp) = split(/:/,$x[2],2); # class no.
		my @res = @{$is[$i]}; shift @res; push @res,$cl;
		print F join("\t",@res),"\n";
		$i++;
	}
	close(G);
	unlink "train-wv.arff";
	unlink "$$.test.arff";
	unlink "$$.test-wv.arff";

	@is=();

  sleep(55);
}

# process rest of old file, close it and restart
Net::Pcap::dispatch($capt, 0, \&process_pkt, "TCP");
Net::Pcap::close($capt);
close(F);

my @args = ($0, $fn2,'-all');
exec { $args[0] } @args;

sub process_pkt_ignore { }

sub process_pkt {
  my ($user_data, $hdr, $pkt) = @_;

  my $pkt_id = $fn."-".$hdr->{"tv_sec"}."-".$hdr->{"tv_usec"};

  my $eth_obj = NetPacket::Ethernet->decode($pkt);
  if ($eth_obj->{type} == NetPacket::Ethernet::ETH_TYPE_IP) {
    my $ip_obj = NetPacket::IP->decode($eth_obj->{data});
    # foreach my $k (keys %{$ip_obj}) { print $k,"=",$ip_obj->{$k},"\n"; }
    $pkt_id.="-".$ip_obj->{proto}."-".$ip_obj->{id};
    my @p = split(/\./,$ip_obj->{dest_ip}); my $ok=undef;
    my $offs=0;
    if ($p[0] == 131 && $p[1] == 130) {
      if ($p[2] == 42 && $p[3]>=192 && $p[3]<=255) { $ok="A"; $offs = $p[3]-192; }
      elsif ($p[2] == 67 && $p[3]>=192 && $p[3]<=255) { $ok="B"; $offs = $p[3]-192; }
      elsif ($p[2] == 167 && $p[3]>=128 && $p[3]<=191) { $ok="C"; $offs = $p[3]-128; }
      elsif ($p[2] == 182 && $p[3]>=0 && $p[3]<=63) { $ok="D"; $offs = $p[3]; }
     }
     if (defined($ok)) { # from our subnets
       my $DNSBL=undef;
       if (defined($dnscache{$ip_obj->{src_ip}})) { $DNSBL=$dnscache{$ip_obj->{src_ip}}; } else {
         my @q = split(/\./,$ip_obj->{src_ip}); @q = reverse @q;
         my $query = $res->search(join(".",@q).".zen.spamhaus.org");
         if ($query) {
           my @r=();
           foreach my $rr ($query->answer) {
             next unless $rr->type eq "A";
             if ($rr->address eq "127.0.0.2") { push @r, "SBL"; }
             elsif ($rr->address eq "127.0.0.4") { push @r, "XBL-CBL"; }
             elsif ($rr->address eq "127.0.0.5") { push @r, "XBL-NJABL"; }
             elsif ($rr->address eq "127.0.0.6") { push @r, "XBL-OTHER-6"; }
             elsif ($rr->address eq "127.0.0.7") { push @r, "XBL-OTHER-7"; }
             elsif ($rr->address eq "127.0.0.8") { push @r, "XBL-OTHER-8"; }
             elsif ($rr->address eq "127.0.0.10") { push @r, "PBL-ISP"; }
             elsif ($rr->address eq "127.0.0.11") { push @r, "PBL-SPAMHAUS"; }
             else { push @r, "UNKNOWN-".$rr->address; }
           }
           @r = sort { $a cmp $b } @r;
           if (@r==0) { $DNSBL="UNKNWON-NO_ENTRY"; } else { $DNSBL=join(",",@r); }
         } else { $DNSBL="?"; }
         $dnscache{$ip_obj->{src_ip}}=$DNSBL;
       }
			 my @inst = ();

				push @inst,($ip_obj->{proto} == 1 ? 1 : 0); # isICMP
				push @inst,($ip_obj->{proto} == 6 ? 1 : 0); # isTCP
				push @inst,($ip_obj->{flags} & 2 == 2 ? 1 : 0); # hasDontFragment

				my $data;
				if ($ip_obj->{proto} == 1) {
					my $icmp = NetPacket::ICMP->decode($ip_obj->{data});
					$data=$icmp->{data}; push @inst,"ICMP";
				} elsif ($ip_obj->{proto} == 6) {
					my $tcp = NetPacket::TCP->decode($ip_obj->{data});
					$data=$tcp->{options}; my $p = $tcp->{dest_port};
					if ($legal{$p}) { push @inst, $p; } else { push @inst, "UNKNOWN_PORT"; }
				} else {
					push @inst, "UNKNOWN_PROTOCOL"; $data="";
				}

				my $d = $data;
				my $str=""; my $ngS=2; # 2grams
				for (my $k=0; $k<length($d)-$ngS; $k++) {
				  my $x = substr($d,$k,$ngS);
				  $str .= sprintf("%02x%02x",ord(substr($x,0,1)),ord(substr($x,1,1))).",";
				}
				$str = substr($str,0,length($str)-1);
				$str = '"'.$str.'"';
				push @inst,$str;

				push @inst,'?'; # class

				push @is,[\@inst,$pkt_id,$ip_obj->{src_ip},$ok.$offs,$DNSBL];
     }
  }
}
